<?php
namespace nur\sery\file\csv;

use DateTimeInterface;
use nur\sery\cl;
use nur\sery\file\TempStream;
use nur\sery\php\func;
use nur\sery\php\time\DateTime;
use nur\sery\ValueException;
use nur\sery\web\http;

/**
 * Class CsvBuilder: construction d'un fichier CSV, pour envoi à l'utilisateur
 */
class CsvBuilder extends TempStream {
  static function with($builder): self {
    if ($builder instanceof self) return $builder;
    elseif (is_string($builder)) return new static($builder);
    elseif (is_array($builder)) return new static(null, $builder);
    else throw ValueException::invalid_type($builder, self::class);
  }

  /** @var ?array schéma des données à écrire */
  const SCHEMA = null;

  /** @var ?array liste des colonnes à écrire */
  const HEADERS = null;

  /** @var ?string nom du fichier téléchargé */
  const OUTPUT = null;

  function __construct(?string $output, ?array $params=null) {
    if ($output !== null) $params["output"] = $output;
    $csvFlavour = $params["csv_flavour"] ?? null;
    $this->csvFlavour = csv_flavours::verifix($csvFlavour);
    $this->schema = $params["schema"] ?? static::SCHEMA;
    $this->headers = $params["headers"] ?? static::HEADERS;
    $this->rows = $params["rows"] ?? null;
    $cookFunc = $params["cook_func"] ?? null;
    $cookCtx = $cookArgs = null;
    if ($cookFunc !== null) {
      func::ensure_func($cookFunc, $this, $cookArgs);
      $cookCtx = func::_prepare($cookFunc);
    }
    $this->cookCtx = $cookCtx;
    $this->cookArgs = $cookArgs;
    $this->output = $params["output"] ?? static::OUTPUT;
    $maxMemory = $params["max_memory"] ?? null;
    $throwOnError = $params["throw_on_error"] ?? null;
    parent::__construct($maxMemory, $throwOnError);
  }

  protected ?array $schema;

  protected ?array $headers;

  protected ?iterable $rows;

  protected ?string $output;

  protected ?array $cookCtx;

  protected ?array $cookArgs;

  protected function ensureHeaders(?array $row=null): void {
    if ($this->headers !== null) return;
    if ($this->schema === null) $headers = null;
    else $headers = array_keys($this->schema);
    if ($headers === null && $row !== null) $headers = array_keys($row);
    $this->headers = $headers;
  }

  protected bool $csvHeadersSent = false;

  function writeHeaders(?array $headers=null): void {
    if ($this->csvHeadersSent) return;
    if ($headers !== null) $this->headers = $headers;
    else $this->ensureHeaders();
    if ($this->headers !== null) $this->fputcsv($this->headers);
    $this->csvHeadersSent = true;
  }

  protected function cookRow(?array $row): ?array {
    if ($this->cookCtx !== null) {
      $args = cl::merge([$row], $this->cookArgs);
      $row = func::_call($this->cookCtx, $args);
    }
    if ($row !== null) {
      foreach ($row as &$value) {
        # formatter les dates
        if ($value instanceof DateTime) {
          $value = $value->format();
        } elseif ($value instanceof DateTimeInterface) {
          $value = DateTime::with($value)->format();
        }
      }; unset($value);
    }
    return $row;
  }

  function write(?array $row): void {
    $row = $this->cookRow($row);
    if ($row === null) return;
    $this->ensureHeaders($row);
    $this->fputcsv($row);
  }

  function writeAll(?iterable $rows=null): void {
    $unsetRows = false;
    if ($rows === null) {
      $rows = $this->rows;
      $unsetRows = true;
    }
    if ($rows !== null) {
      foreach ($rows as $row) {
        $this->write($row);
      }
    }
    if ($unsetRows) $this->rows = null;
  }

  protected bool $httpHeadersSent = false;

  function sendHeaders(): void {
    if ($this->httpHeadersSent) return;
    http::content_type("text/csv");
    $output = $this->output;
    if ($output !== null) http::download_as($output);
    $this->httpHeadersSent = true;
  }

  function sendFile(?iterable $rows=null): int {
    $this->writeAll($rows);
    $this->writeHeaders();
    $size = $this->ftell();
    if ($size === 0) return 0;
    $this->rewind();
    $this->sendHeaders();
    return $this->fpassthru();
  }
}
